<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.2/1705499
 */

namespace Workflow;

use Application\Config\ConfigException;
use Application\Config\ConfigManager;
use Application\Permissions\Exception\ForbiddenException;
use P4\Connection\Connection;
use P4\Exception;
use P4\Log\Logger;
use P4\Spec\Change;
use P4\Spec\Exception\NotFoundException;
use Projects\Model\Project;
use Reviews\Model\Review;
use Workflow\Model\IWorkflow;
use Workflow\Model\Workflow;
use Zend\ServiceManager\ServiceLocatorInterface as ServiceLocator;
use Record\Exception\NotFoundException as RecordNotFoundException;

/**
 * Class to manage workflow decisions.
 * @package Workflow
 */
class Manager
{
    const KEY_STATUS                = 'status';
    const KEY_MESSAGES              = 'messages';
    const STATUS_OK                 = 'OK';
    const STATUS_NO_REVIEW          = 'NO_REVIEW';
    const STATUS_NO_APPROVED_REVIEW = 'NO_APPROVED_REVIEW';
    const STATUS_BAD_CHANGE         = 'BAD_CHANGE';
    const STATUS_CREATED_REVIEW     = 'CREATED_REVIEW';
    const STATUS_LINKED_REVIEW      = 'LINKED_REVIEW';
    const STATUS_NO_REVISION        = 'NO_REVISION';
    const STATUS_NOT_SAME_CONTENT   = 'NOT_SAME_CONTENT';
    const BAD_CHANGE_ID             = 'Must supply a valid id to fetch.';

    private $messages;
    private $logger;
    private $services                     = null;
    private $vanilla                      = array(
        IWorkflow::WORKFLOW_RULES => array(
            IWorkflow::ON_SUBMIT => array(
                IWorkflow::WITH_REVIEW    => array(IWorkflow::RULE => IWorkflow::NO_CHECKING),
                IWorkflow::WITHOUT_REVIEW => array(IWorkflow::RULE => IWorkflow::NO_CHECKING),
            ),
            IWorkflow::END_RULES => array(
                IWorkflow::UPDATE => array(IWorkflow::RULE => IWorkflow::NO_CHECKING)
            ),
            IWorkflow::AUTO_APPROVE => array(IWorkflow::RULE => IWorkflow::NEVER),
            IWorkflow::COUNTED_VOTES => array(IWorkflow::RULE => IWorkflow::ANYONE)
        ));
    private static $workflowRulesTemplate = array(
        IWorkflow::ON_SUBMIT => array(
            IWorkflow::WITH_REVIEW    => array(IWorkflow::RULE => null),
            IWorkflow::WITHOUT_REVIEW => array(IWorkflow::RULE => null),
        ),
        IWorkflow::END_RULES => array(
            IWorkflow::UPDATE => array(IWorkflow::RULE => null)
        ),
        IWorkflow::AUTO_APPROVE => array(IWorkflow::RULE => null),
        IWorkflow::COUNTED_VOTES => array(IWorkflow::RULE => null)
    );


    /**
     * Constructs the manager
     * @param ServiceLocator $services the service locator to use
     */
    public function __construct(ServiceLocator $services)
    {
        $this->services = $services;
        $translator     = $services->get('translator');
        $this->logger   = $services->get('logger');
        $this->messages = array(
            self::STATUS_NO_REVIEW          => $translator->t('Change [%s] must be associated with a review'),
            self::STATUS_NO_APPROVED_REVIEW => $translator->t('Review [%s] must be approved'),
            self::STATUS_CREATED_REVIEW     => $translator->t('Review [%s] created for change [%s]'),
            self::STATUS_LINKED_REVIEW      => $translator->t('Review [%s] linked to change [%s]'),
            self::STATUS_NO_REVISION        => $translator->t('Review [%s] is in state [%s] and cannot be updated'),
            self::STATUS_NOT_SAME_CONTENT   => $translator->t('Change [%s] content is different from Review [%s]')
        );
    }

    /**
     * Gets a message for the key.
     * @param string    $key    key for the message
     * @return mixed|null
     */
    private function getMessage($key)
    {
        return isset($this->messages[$key]) ? $this->messages[$key] : '';
    }

    /**
     * Carries out strict checks on the change if enabled
     * @param $changeId string the change id.
     * @return array Containing a status and an array (possibly empty) of messages
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    public function checkStrict($changeId)
    {
        $enforce = $this->enforceWorkflow();
        return $enforce ? $enforce : $this->processEnforcedOrStrict($changeId, true);
    }

    /**
     * Carries out enforced checks on the change if enabled
     * @param $changeId string the change id.
     * @return array
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    public function checkEnforced($changeId)
    {
        $enforce = $this->enforceWorkflow();
        return $enforce ? $enforce : $this->processEnforcedOrStrict($changeId);
    }

    /**
     * Carries out shelve checks on the change if enabled (shelve-submit)
     * @param $changeId string the change id.
     * @return array
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    public function checkShelve($changeId)
    {
        $enforce = $this->enforceWorkflow();
        return $enforce ? $enforce : $this->processShelve($changeId);
    }

    /**
     * Checks whether workflows are enabled and just returns an 'OK' result if they are not.
     * @return array|null
     */
    private function enforceWorkflow()
    {
        try {
            $this->services->get('config_check')->enforce(IWorkflow::WORKFLOW);
        } catch (ForbiddenException $fe) {
            $this->logger->info($fe->getMessage());
            return $this->buildResult(self::STATUS_OK, $fe->getMessage());
        }
        return null;
    }

    /**
     * Get the rule value associated with the given id based upon the workflows associated with the projects/branches
     * provided, merged with any defaults and policies defined in the swarm config.
     *
     * If the project definition cannot be found in the optional list of projects it will be retrieved using the
     * given connection.
     *
     * @param string         $ruleId          - The unique name of the IWorkflow::RULE values to be merged
     * @param array          $projectBranches - The project=>branch ids that should be merged
     * @param ServiceLocator $services        - The Services to allow us to call Enforce.
     * @param null           $projects        - An optional pre-populated list of project models, missing data will be
     *                                          fetched
     *
     * @return mixed|null      - The net IWorkflow::RULE value, or null if there is no relevant value
     * @throws ConfigException
     * @throws Exception
     */
    public static function getBranchRule($ruleId, $projectBranches, $services, $projects = null)
    {
        $logger    = Logger::getLogger();
        $workflows = array();

        $connection = $services->get('p4_admin');
        $config     = $services->get('config');
        try {
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW);
            $logger->info("Workflow/Manager: Getting '$ruleId' rule for ['" . var_export($projectBranches, true) . "]");
            $projectArray = (array) $projects;
            foreach ($projectBranches as $projectId => $branches) {
                try {
                    if (!isset($projectArray[ $projectId ])) {
                        $projectArray[ $projectId ] = Project::fetch($projectId, $connection);
                    }
                    $workflows = array_merge(
                        $workflows,
                        array_map(
                            function ($branch) use ($projectArray, $projectId) {
                                return $projectArray[ $projectId ]->getWorkflow($branch);
                            },
                            $branches
                        )
                    );
                } catch (RecordNotFoundException $nfe) {
                    $logger->notice("Workflow/Manager: Project [$projectId] is not found, falling back to defaults.");
                }
            }
        } catch (ForbiddenException $fe) {
            // Workflow is disabled
            $logger->debug('Workflow/Manager: getBranchRule disregarding project workflow ' . $fe->getMessage());
        }
        // Merge the list of workflows
        $rule = null;
        foreach (Workflow::fetchAll(
            array(Workflow::FETCH_BY_IDS => array_unique($workflows)),
            $connection
        ) as $workflow) {
            $logger->debug("Workflow/Manager: Processing " . $workflow->getId());
            $rule = static::mergeRule($ruleId, $rule, $workflow);
        }
        // Now apply any global override
        $rule = static::mergeConfigRule($ruleId, $rule, $config);
        return $rule;
    }

    /**
     * Merge the current value of a rule with that from the workflow provided.
     *
     * @param $ruleId       - The unique name of the IWorkflow::RULE value to be merged
     * @param $currentValue - The current value for that rule
     * @param $workflow     - The workflow with a new potential value
     * @return mixed        - The result of the merging, typically a string
     * @throws Exception
     */
    private static function mergeRule($ruleId, $currentValue, $workflow)
    {
        $mergeTo = array(IWorkflow::WORKFLOW_RULES=>static::$workflowRulesTemplate);
        switch ($ruleId) {
            case IWorkflow::AUTO_APPROVE:
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::AUTO_APPROVE][IWorkflow::RULE] = $currentValue;
                static::mergeAutoApproveRule($workflow, $mergeTo);
                $mergedValue = $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::AUTO_APPROVE][IWorkflow::RULE];
                break;
            case IWorkflow::COUNTED_VOTES:
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::COUNTED_VOTES][IWorkflow::RULE] = $currentValue;
                static::mergeCountedVotesRule($workflow, $mergeTo);
                $mergedValue = $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::COUNTED_VOTES][IWorkflow::RULE];
                break;
            case IWorkflow::UPDATE:
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES][IWorkflow::UPDATE]
                    [IWorkflow::RULE] = $currentValue;
                static::mergeUpdateRule($workflow, $mergeTo);
                $mergedValue = $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES][IWorkflow::UPDATE]
                    [IWorkflow::RULE];
                break;
            case IWorkflow::WITH_REVIEW:
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW]
                    [IWorkflow::RULE] = $currentValue;
                static::mergeWithReviewRule($workflow, $mergeTo);
                $mergedValue = $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW]
                    [IWorkflow::RULE];
                break;
            case IWorkflow::WITHOUT_REVIEW:
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW]
                    [IWorkflow::RULE] = $currentValue;
                static::mergeWithoutReviewRule($workflow, $mergeTo);
                $mergedValue = $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW]
                    [IWorkflow::RULE];
                break;
            default:
                throw new Exception("Manager/mergeRule: Unsupported rule id[$ruleId]");
                break;
        }
        return $mergedValue;
    }

    /**
     * Merge the current value of a rule with those in the swarm config, taking account of defaults and policy
     * settings. A policy will always be merged, a default will apply if there is no current value.
     *
     * @param $ruleId       - The unique name of the IWorkflow::RULE value to be merged
     * @param $currentValue - The current value for that rule
     * @param $config       - The configuration object, containing workflow default/policy values
     * @return mixed        - The result of the merging, typically a string
     * @throws ConfigException
     * @throws Exception
     * @throws \Exception
     */
    private static function mergeConfigRule($ruleId, $currentValue, $config)
    {
        $mergedValue    = $currentValue;
        $globalWorkflow = new Workflow();
        $globalWorkflow->set(static::$workflowRulesTemplate);
        switch ($ruleId) {
            case IWorkflow::AUTO_APPROVE:
                $ruleDefaults = static::getGlobalAutoApprove($config);
                if ($ruleDefaults[IWorkflow::MODE] === IWorkflow::MODE_POLICY) {
                    $globalWorkflow->setAutoApprove($ruleDefaults);
                    $mergedValue = static::mergeRule($ruleId, $currentValue, $globalWorkflow);
                }
                break;
            case IWorkflow::COUNTED_VOTES:
                $ruleDefaults = static::getGlobalCountedVotes($config);
                if ($ruleDefaults[IWorkflow::MODE] === IWorkflow::MODE_POLICY) {
                    $globalWorkflow->setCountedVotes($ruleDefaults);
                    $mergedValue = static::mergeRule($ruleId, $currentValue, $globalWorkflow);
                }
                break;
            case IWorkflow::UPDATE:
                $defaultEndRules = static::getGlobalEndRules($config);
                $ruleDefaults    = $defaultEndRules[IWorkflow::UPDATE];
                if ($defaultEndRules[IWorkflow::UPDATE][IWorkflow::MODE] === IWorkflow::MODE_POLICY) {
                    $globalWorkflow->setEndRules($defaultEndRules);
                    $mergedValue = static::mergeRule($ruleId, $currentValue, $globalWorkflow);
                }
                break;
            case IWorkflow::WITH_REVIEW:
                $defaultOnSubmit = static::getGlobalOnSubmit($config);
                $ruleDefaults    = $defaultOnSubmit[IWorkflow::WITH_REVIEW];
                if ($defaultOnSubmit[IWorkflow::WITH_REVIEW][IWorkflow::MODE] === IWorkflow::MODE_POLICY) {
                    $globalWorkflow->setOnSubmit($defaultOnSubmit);
                    $mergedValue = static::mergeRule($ruleId, $currentValue, $globalWorkflow);
                }
                break;
            case IWorkflow::WITHOUT_REVIEW:
                $defaultOnSubmit = static::getGlobalOnSubmit($config);
                $ruleDefaults    = $defaultOnSubmit[IWorkflow::WITHOUT_REVIEW];
                if ($defaultOnSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::MODE] === IWorkflow::MODE_POLICY) {
                    $globalWorkflow->setOnSubmit($defaultOnSubmit);
                    $mergedValue = static::mergeRule($ruleId, $currentValue, $globalWorkflow);
                }
                break;
            default:
                throw new Exception("Manager/mergeRule: Unsupported rule id[$ruleId]");
                break;
        }
        if ($mergedValue === null) {
            return $ruleDefaults[IWorkflow::RULE];
        }
        return $mergedValue;
    }

    /**
     * Checks on shelve-submit to only permit if there is either no review or there is a review and we are
     * allowing a change based on its state (end_states)
     * @param $changeId string the change id.
     * @return array
     * @throws ConfigException
     * @throws Exception
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    private function processShelve($changeId)
    {
        // Validate the change id early. We don't want to continue and query for reviews if the
        // change is not known
        $this->logger->trace(
            get_class($this) . ": Process workflow for change (shelve) " . var_export($changeId, true)
        );
        $result = $this->validateChangeId($changeId);
        if ($result[self::KEY_STATUS] === self::STATUS_OK) {
            $this->logger->trace(get_class($this) . ": Change $changeId is valid");
            $review = $this->getReview($changeId);
            $this->logger->trace(
                get_class($this) . ": Review for change $changeId is " . ($review ? $review->getId() : 'not found')
            );
            // It may not yet be linked to a review but # processing for a description may be going to link it later so
            // we need to find that review if it is the case and see what state that is in
            if ($review === null) {
                $review = $this->getLinkedReview($result[$changeId]);
            }
            if ($review !== null) {
                // For a shelf with a review we are only interested in workflows with IWorkflow::REJECT
                $workflow = $this->getMergedWorkflow(
                    $result[$changeId],
                    $review->getProjects(),
                    array(IWorkflow::REJECT)
                );
                $this->logger->trace(
                    get_class($this) . ": Merged workflow in process shelve " . var_export($workflow, true)
                );
                $result = $this->checkEndRules($review, $workflow);
            }
        }
        unset($result[$changeId]);
        return $result;
    }

    /**
     * Checks whether the review obeys the end rules criteria
     * @param Review    $review     the current review
     * @param IWorkflow $workflow   current merged workflow, if not provided it will be determined from the
     *                              projects on the review
     * @return array|null
     * @throws ConfigException
     * @throws \Exception
     */
    public function checkEndRules($review, $workflow = null)
    {
        $result = null;

        $workflow = $workflow ? $workflow : $this->getMergedWorkflow(null, $review->getProjects());
        $rule     = $workflow[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES]
                             [IWorkflow::UPDATE][IWorkflow::RULE];
        if ($rule === IWorkflow::NO_REVISION) {
            $endStates = ConfigManager::getValue(
                $this->services->get('config'),
                ConfigManager::REVIEWS_END_STATES
            );
            $this->logger->trace(
                get_class($this) .
                sprintf(
                    " Check end rules, review state [%s], commit count [%s]",
                    $review->getState(),
                    count($review->getCommits())
                )
            );
            $invalid = in_array($review->getState(), $endStates);
            // When the review state is in the endStates
            if ($invalid || in_array(
                $review->getState(),
                array_map(
                    function ($entry) {
                        return strpos($entry, ":") !== false ? strstr($entry, ":", true) : $entry;
                    },
                    $endStates
                )
            )) {
                // Specific case for 'approved' end state and performing an 'approve and commit' from the UI.
                // The state is changed first and we need to avoid erroring on the state of 'approved' not
                // being able to be updated
                if (Review::STATE_APPROVED === $review->getState()
                    && $review->getCommitStatus('status') === 'Committing') {
                    return $this->buildResult(self::STATUS_OK);
                }
                // Allow for approved:commit being in the end states
                if ($invalid || !in_array(Review::STATE_APPROVED_COMMIT, $endStates) ||
                    !Review::STATE_APPROVED === $review->getState() ||
                    count($review->getCommits()) !== 0) {
                    $result = $this->buildResult(
                        self::STATUS_NO_REVISION,
                        sprintf($this->getMessage(self::STATUS_NO_REVISION), $review->getId(), $review->getState())
                    );
                }
            }
        }
        return $result ? $result : $this->buildResult(self::STATUS_OK);
    }

    /**
     * Processes the rule
     * @param string        $changeId     the change id
     * @param bool          $checkContent If we want to check the content of the changelist.
     * @return array
     * @throws ConfigException
     * @throws Exception
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    private function processEnforcedOrStrict($changeId, $checkContent = false)
    {
        // Validate the change id early. We don't want to continue and query for reviews if the
        // change is not known
        $this->logger->trace(
            get_class($this) . ": Process workflow for change (enforced) " . var_export($changeId, true)
        );
        $result = $this->validateChangeId($changeId);
        if ($result[self::KEY_STATUS] === self::STATUS_OK) {
            $this->logger->trace(get_class($this) . ": Change $changeId is valid");
            $review = $this->getReview($changeId);
            if ($review === null) {
                // It may not yet be linked to a review but # processing for a description may be going to link it later
                // so we need to find that review if it is the case and see what state that is in
                $review = $this->getLinkedReview($result[$changeId]);
            }
            $this->logger->trace(
                get_class($this) . ": Review for change $changeId is " . ($review ? $review->getId() : 'not found')
            );
            $workflow = $this->getMergedWorkflow(
                $result[$changeId],
                $review === null ? null : $review->getProjects(),
                $review === null
                    ? (!$checkContent
                        ? array(IWorkflow::NO_CHECKING, IWorkflow::AUTO_CREATE, IWorkflow::REJECT)
                        : array())
                    : array(IWorkflow::NO_CHECKING, IWorkflow::APPROVED, IWorkflow::NO_REVISION)
            );
            $this->logger->trace(get_class($this) . ": Merged workflow in process rule " . var_export($workflow, true));
            if ($review === null) {
                $rule = $workflow[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE];
            } else {
                $rule = $workflow[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                [IWorkflow::WITH_REVIEW][IWorkflow::RULE];
            }
            $this->logger->trace(get_class($this) . ": Rule is $rule");
            if ($review !== null) {
                switch ($rule) {
                    case IWorkflow::APPROVED:
                    case IWorkflow::STRICT:
                        // On submit any # processing will have happened so the change will already be linked
                        // if the description specified so we do not need to look at the change description in this case
                        $result = $this->checkEndRules($review, $workflow, $result[$changeId]);
                        if ($result[self::KEY_STATUS] === self::STATUS_OK) {
                            // If checkContent is true we shouldn't need to re check required Approved
                            if ($checkContent) {
                                $result = $this->checkContent($review, $changeId);
                            } else {
                                $result = $this->requireApproved($review);
                            }
                        }
                        break;
                    case IWorkflow::NO_CHECKING:
                        // Even if we aren't checking for approved we still need to make
                        // sure we are not violating the end rule/end state configuration
                        $result = $this->checkEndRules($review, $workflow);
                        break;
                }
            } else {
                switch ($rule) {
                    case IWorkflow::NO_CHECKING:
                        break;
                    case IWorkflow::AUTO_CREATE:
                        $result = $this->autoCreateReview($result[$changeId]);
                        break;
                    case IWorkflow::REJECT:
                        $result = $this->requireReview($changeId);
                        break;
                }
            }
            unset($result[$changeId]);
        }
        return $result;
    }

    /**
     * Gets a review if it exists
     * @param string        $changeId   the change id
     * @return null|\P4\Model\Connected\ModelInterface
     */
    private function getReview($changeId)
    {
        $p4Admin = $this->services->get('p4_admin');
        $review  = null;
        $reviews = Review::fetchAll(array(Review::FETCH_BY_CHANGE => $changeId), $p4Admin);
        $this->logger->trace(get_class($this) . ": Found " . $reviews->count() . " review(s) for change $changeId");
        if ($reviews->count() !== 0) {
            $review = $reviews->first();
        }
        return $review;
    }

    /**
     * Check that the users changelist and the head changelist of the review content is the same.
     * We can let the P4D server do the checking for us as diff2 check the content without expanding
     * the keywords
     *
     * @param Review $review
     * @param int    $changeId
     *
     * @return array result
     */
    private function checkContent($review, $changeId)
    {
        // Get the admin connection and then use diff2 to check the content diff.
        // Diff2 handles ktext file digest.
        $p4        = $this->services->get('p4_admin');
        $flags     = array('-ds', '//...@='.$review->getHeadChange(), '//...@='.$changeId);
        $diff2     = $p4->run('diff2', $flags);
        $diff2Data = $diff2->getData();
        // Loop though each of the files and if we find one is not identical break early.
        foreach ($diff2Data as $file) {
            if (isset($file['status']) && $file['status'] !== 'identical') {
                return $this->buildResult(
                    self::STATUS_NOT_SAME_CONTENT,
                    sprintf($this->getMessage(self::STATUS_NOT_SAME_CONTENT), $changeId, $review->getId())
                );
            }
        }
        return $this->buildResult(self::STATUS_OK);
    }

    /**
     * Enforces that the review is approved
     * @param Review $review the review.
     * @return array result
     */
    private function requireApproved($review)
    {
        $this->logger->trace(get_class($this) . ": In requireApproved, review state is " . $review->getState());
        if ($review->getState() === Review::STATE_APPROVED) {
            return $this->buildResult(self::STATUS_OK);
        } else {
            return $this->buildResult(
                self::STATUS_NO_APPROVED_REVIEW,
                sprintf($this->getMessage(self::STATUS_NO_APPROVED_REVIEW), $review->getId())
            );
        }
    }

    /**
     * Enforces auto create by creating a review if it does not already exist.
     * @param $change Change the change.
     * @return array result of the action
     * @throws \Record\Exception\NotFoundException
     */
    private function autoCreateReview($change)
    {
        $changeId   = $change->getId();
        $p4Admin    = $this->services->get('p4_admin');
        $review     = $this->getLinkedReview($change);
        $messageKey = self::STATUS_CREATED_REVIEW;
        // We only want to auto-create a review if the change wasn't linked to an existing
        // review by #review #append etc
        if ($review) {
            $messageKey = self::STATUS_LINKED_REVIEW;
        } else {
            $review = Review::createFromChange($changeId, $p4Admin)->save();
        }
        $message = sprintf($this->getMessage($messageKey), $review->getId(), $changeId);
        return $this->buildResult(self::STATUS_OK, $message);
    }

    /**
     * Checks the description of the change to see if it will be linked to a review and returns that
     * review if found.
     * @param $change
     * @return null|\Record\Key\AbstractKey
     * @throws \Record\Exception\NotFoundException
     */
    private function getLinkedReview($change)
    {
        $review      = null;
        $p4Admin     = $this->services->get('p4_admin');
        $keywords    = $this->services->get('review_keywords');
        $description = $change->getDescription();
        $this->logger->trace(get_class($this) . " Description in getLinkedReview " . $description);
        $matches = $keywords->getMatches($description);
        if ($matches && isset($matches['id']) && !empty($matches['id'])) {
            $this->logger->trace(get_class($this) . ": Matches from change description " . var_export($matches, true));
            try {
                $review = Review::fetch($matches['id'], $p4Admin);
            } catch (RecordNotFoundException $e) {
                // This is OK #review-x has been provided but x does not relate to a review
                $this->logger->trace(get_class($this) . ": Review from matches was not found");
            }
        }
        return $review;
    }

    /**
     * Enforces that the provided change has a review
     * @param $changeId string the change id.
     * @return array
     */
    private function requireReview($changeId)
    {
        return $this->buildResult(
            self::STATUS_NO_REVIEW,
            sprintf($this->getMessage(self::STATUS_NO_REVIEW), $changeId)
        );
    }

    /**
     * Gets the 'end_rules' settings as defined in the global configuration
     * @return array
     * @throws \Application\Config\ConfigException
     */
    private static function getGlobalEndRules($config)
    {
        return array(
            IWorkflow::UPDATE => array(
                IWorkflow::RULE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_END_RULES_UPDATE_RULE,
                        IWorkflow::NO_CHECKING
                    ),
                IWorkflow::MODE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_END_RULES_UPDATE_MODE,
                        IWorkflow::MODE_DEFAULT
                    )
            )
        );
    }

    /**
     * Gets the 'auto_approve' settings as defined in the global configuration
     * @return array
     * @throws \Application\Config\ConfigException
     */
    private static function getGlobalAutoApprove($config)
    {
        return array(
            IWorkflow::RULE =>
                ConfigManager::getValue(
                    $config,
                    ConfigManager::WORKFLOW_AUTO_APPROVE_RULE,
                    IWorkflow::NEVER
                ),
            IWorkflow::MODE =>
                ConfigManager::getValue(
                    $config,
                    ConfigManager::WORKFLOW_AUTO_APPROVE_MODE,
                    IWorkflow::MODE_DEFAULT
                )
        );
    }

    /**
     * Gets the 'on_submit' settings as defined in the global configuration
     * @return array
     * @throws \Application\Config\ConfigException
     */
    private static function getGlobalOnSubmit($config)
    {
        return array(
            IWorkflow::WITH_REVIEW => array(
                IWorkflow::RULE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_ON_SUBMIT_WITH_REVIEW_RULE,
                        IWorkflow::NO_CHECKING
                    ),
                IWorkflow::MODE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_ON_SUBMIT_WITH_REVIEW_MODE,
                        IWorkflow::MODE_DEFAULT
                    )
            ),
            IWorkflow::WITHOUT_REVIEW => array(
                IWorkflow::RULE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_RULE,
                        IWorkflow::NO_CHECKING
                    ),
                IWorkflow::MODE =>
                    ConfigManager::getValue(
                        $config,
                        ConfigManager::WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_MODE,
                        IWorkflow::MODE_DEFAULT
                    )
            )
        );
    }

    /**
     * Gets the 'counted_vote' settings as defined in the global configuration
     * @return array
     * @throws \Application\Config\ConfigException
     * @throws \Exception
     */
    private static function getGlobalCountedVotes($config)
    {
        return array(
            IWorkflow::RULE =>
                ConfigManager::getValue(
                    $config,
                    ConfigManager::WORKFLOW_COUNTED_VOTES_RULE,
                    IWorkflow::ANYONE
                ),
            IWorkflow::MODE =>
                ConfigManager::getValue(
                    $config,
                    ConfigManager::WORKFLOW_COUNTED_VOTES_MODE,
                    IWorkflow::MODE_DEFAULT
                )
        );
    }

    /**
     * Looks at the impacted projects and branches and returns merged workflow settings with the most restrictive
     * setting winning. If there are no workflows defined for any of the impacted projects or branches null is returned.
     * @param array     $impacted       impacted projects
     * @return array|null
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    // Avoids a potentially expensive call to Project::getAffectedProjects for when projects are known.
    public function getMergedAffectedWorkflowRules($impacted)
    {
        $mergeTo           = $this->vanilla;
        $hasBranchWorkflow = false;
        $p4Admin           = $this->services->get('p4_admin');
        $projectWf         = null;
        $this->logger->trace(get_class($this) . ": Impacted projects for change " . var_export($impacted, true));
        foreach ($impacted as $projectId => $branches) {
            $project    = Project::fetch($projectId, $p4Admin);
            $workflowId = $project->getWorkflow();
            if ($workflowId) {
                $this->logger->trace(get_class($this) . ": Processing workflow $workflowId for project $projectId");
                try {
                    $projectWf = Workflow::fetch($workflowId, $p4Admin);
                } catch (RecordNotFoundException $e) {
                    $this->logger->warn(get_class($this) . ": Workflow $workflowId for project $projectId not found");
                }
            }
            foreach ($branches as $branch) {
                $workflowId = $project->getWorkflow($branch);
                // It will be the project level workflow returned if the branch is invalid or does
                // not have its own workflow. We only need to merge if the branch has provided its own
                if ($workflowId && ($projectWf === null || $workflowId !== $projectWf->getId())) {
                    $this->logger->trace(
                        get_class($this) . ": Processing workflow $workflowId for project $projectId branch $branch"
                    );
                    try {
                        $this->mergeWorkflow(Workflow::fetch($workflowId, $p4Admin), $mergeTo);
                        $hasBranchWorkflow = true;
                    } catch (RecordNotFoundException $e) {
                        $this->logger->warn(
                            get_class($this) . ": Workflow $workflowId for branch $branch not found"
                        );
                    }
                }
            }
            // Only use the project level workflow if there were none on branches
            if (!$hasBranchWorkflow && $projectWf) {
                $this->mergeWorkflow($projectWf, $mergeTo);
            }
        }
        return $hasBranchWorkflow || $projectWf != null ? $mergeTo : null;
    }

    /**
     * Looks at the affected projects and branches for the change and returns merged workflow
     * settings with the most restrictive setting winning. If there are no workflows defined for any
     * of the impacted projects or branches null is returned.
     * @param Change    $change     the change spec
     * @param $rulesToCheck which workflow rules are being merged, which allows projects to be limited by workflow
     * @return array|null
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    private function getMergedChangeWorkflowRules($change, $rulesToCheck = null)
    {
        $p4Admin  = $this->services->get('p4_admin');
        $projects = null;
        if ($rulesToCheck) {
            // Limit project scope as quickly as possible, by looking for projects with workflows with specific rules
            $projects = array();
            // Find all of the workflows that have the given rules values set
            $keywords    = implode(" ", (array)$rulesToCheck);
            $newWorkflow = new Workflow;
            $workflows   = array_keys(
                (array)Workflow::fetchAll(
                    array(
                        Workflow::FETCH_BY_KEYWORDS => $keywords,
                        Workflow::FETCH_KEYWORDS_FIELDS => array_filter(
                            $newWorkflow->getDefinedFields($p4Admin),
                            function ($field) use ($rulesToCheck) {
                                return $field !== IWorkflow::NAME;
                            }
                        )
                    ),
                    $p4Admin
                )
            );
            if (!empty($workflows)) {
                // There are workflows with these rules, find the associated projects
                $projects = Project::fetchAll(array(Project::FETCH_BY_WORKFLOW => $workflows), $p4Admin);
            }
        }
        // First try to get impacted projects from a shelved (default p4 describe with '-sS')
        $impacted = Project::getAffectedByChange($change, $p4Admin, '-sS', $projects);
        if (empty($impacted)) {
            // If there are no impacted projects try again with p4 describe '-s' as this may
            // have been a straight submit without shelved files
            $impacted = Project::getAffectedByChange($change, $p4Admin, '-s', $projects);
        }
        return $this->getMergedAffectedWorkflowRules($impacted);
    }

    /**
     * Merges the workflow into the rolling array representation with the strictest rule winning
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private function mergeWorkflow(IWorkflow $workflow, &$mergeTo)
    {
        static::mergeWithReviewRule($workflow, $mergeTo);
        static::mergeWithoutReviewRule($workflow, $mergeTo);
        static::mergeUpdateRule($workflow, $mergeTo);
        static::mergeAutoApproveRule($workflow, $mergeTo);
        static::mergeCountedVotesRule($workflow, $mergeTo);
        $this->logger->trace(get_class($this) . ": Merged workflow " . var_export($mergeTo, true));
    }

    /**
     * Merges the workflow 'with_review' into the rolling array representation with the strictest rule winning
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private static function mergeWithReviewRule(IWorkflow $workflow, &$mergeTo)
    {
        $wfOnSubmit = $workflow->getOnSubmit();
        if (isset($wfOnSubmit[IWorkflow::WITH_REVIEW])) {
            $withReviewRule = $wfOnSubmit[IWorkflow::WITH_REVIEW][IWorkflow::RULE];
            switch ($withReviewRule) {
                case IWorkflow::APPROVED:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITH_REVIEW][IWorkflow::RULE] === IWorkflow::NO_CHECKING) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITH_REVIEW][IWorkflow::RULE] = $withReviewRule;
                    }
                    break;
                case IWorkflow::STRICT:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITH_REVIEW][IWorkflow::RULE] === IWorkflow::NO_CHECKING ||
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITH_REVIEW][IWorkflow::RULE] === IWorkflow::APPROVED) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITH_REVIEW][IWorkflow::RULE] = $withReviewRule;
                    }
                    break;
                default:
                    break;
            }
            // There is no value, use whatever is passed in
            if (!isset(
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW][IWorkflow::RULE]
            )) {
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITH_REVIEW]
                [IWorkflow::RULE] = $withReviewRule;
            }
        }
    }

    /**
     * Merges the workflow 'auto_approve' into the rolling array representation with the strictest rule winning.
     * In this case 'NEVER' is more strict
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private static function mergeAutoApproveRule(IWorkflow $workflow, &$mergeTo)
    {
        $wfAutoApprove = $workflow->getAutoApprove();
        if (isset($wfAutoApprove[IWorkflow::RULE])) {
            $rule = $wfAutoApprove[IWorkflow::RULE];
            switch ($rule) {
                case IWorkflow::NEVER:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES]
                        [IWorkflow::AUTO_APPROVE][IWorkflow::RULE] === IWorkflow::VOTES) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES]
                        [IWorkflow::AUTO_APPROVE][IWorkflow::RULE] = $rule;
                    }
                    break;
                default:
                    break;
            }
            // There is no value, use whatever is passed in
            if (!isset($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::AUTO_APPROVE][IWorkflow::RULE])) {
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::AUTO_APPROVE][IWorkflow::RULE] = $rule;
            }
        }
    }

    /**
     * Merges the workflow 'update' into the rolling array representation with the strictest rule winning
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private static function mergeUpdateRule(IWorkflow $workflow, &$mergeTo)
    {
        $wfEndRules = $workflow->getEndRules();
        if (isset($wfEndRules[IWorkflow::UPDATE])) {
            $rule = $wfEndRules[IWorkflow::UPDATE][IWorkflow::RULE];
            switch ($rule) {
                case IWorkflow::NO_REVISION:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES]
                        [IWorkflow::UPDATE][IWorkflow::RULE] === IWorkflow::NO_CHECKING) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES]
                        [IWorkflow::UPDATE][IWorkflow::RULE] = $rule;
                    }
                    break;
                default:
                    break;
            }
            // There is no value, use whatever is passed in
            if (!isset(
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES][IWorkflow::UPDATE][IWorkflow::RULE]
            )) {
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::END_RULES][IWorkflow::UPDATE]
                [IWorkflow::RULE] = $rule;
            }
        }
    }

    /**
     * Merges the workflow 'without_review' into the rolling array representation with the strictest rule winning
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private static function mergeWithoutReviewRule(IWorkflow $workflow, &$mergeTo)
    {
        $wfOnSubmit = $workflow->getOnSubmit();
        if (isset($wfOnSubmit[IWorkflow::WITHOUT_REVIEW])) {
            $withoutReviewRule = $wfOnSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE];
            switch ($withoutReviewRule) {
                case IWorkflow::AUTO_CREATE:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] === IWorkflow::NO_CHECKING) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] = $withoutReviewRule;
                    }
                    break;
                case IWorkflow::REJECT:
                    if ($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] === IWorkflow::NO_CHECKING ||
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] === IWorkflow::AUTO_CREATE) {
                        $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT]
                        [IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE] = $withoutReviewRule;
                    }
                    break;
                default:
                    break;
            }
            // There is no value, use whatever is passed in
            if (!isset(
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW][IWorkflow::RULE]
            )) {
                $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::ON_SUBMIT][IWorkflow::WITHOUT_REVIEW]
                [IWorkflow::RULE] = $withoutReviewRule;
            }
        }
    }

    /**
     * Merges the workflow 'counted_votes' into the rolling array representation with the strictest rule winning. The
     * counted votes rule is very simple members takes precedence.
     * @param IWorkflow     $workflow   the workflow
     * @param array         $mergeTo    the rolling merged workflow
     */
    private static function mergeCountedVotesRule(IWorkflow $workflow, &$mergeTo)
    {
        $countedVotes = $workflow->getCountedVotes();
        if (isset($countedVotes[IWorkflow::RULE]) &&
            (   $countedVotes[IWorkflow::RULE] === IWorkflow::MEMBERS ||
                !isset($mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::COUNTED_VOTES][IWorkflow::RULE]))
        ) {
            $mergeTo[IWorkflow::WORKFLOW_RULES][IWorkflow::COUNTED_VOTES][IWorkflow::RULE] =
                $countedVotes[IWorkflow::RULE];
        }
    }

    /**
     * Builds a merged workflow based on the global settings and any workflows defined on the projects and branches
     * that the change impacts
     * @param Change        $change             the change spec used to find affected projects
     * @param array         $affectedProjects   if provided these projects are used in preference rather than retrieving
     *                                          based on the change
     * @param $rulesToCheck which rules are relevant to the client scope
     * @return array|null
     * @throws ConfigException
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    private function getMergedWorkflow($change, $affectedProjects = null, $rulesToCheck = null)
    {
        $p4Admin           = $this->services->get('p4_admin');
        $config            = $this->services->get('config');
        $onSubmit          = $this->getGlobalOnSubmit($config);
        $endRules          = $this->getGlobalEndRules($config);
        $autoApprove       = $this->getGlobalAutoApprove($config);
        $countedVotes      = $this->getGlobalCountedVotes($config);
        $withReviewMode    = $onSubmit[IWorkflow::WITH_REVIEW][IWorkflow::MODE];
        $withoutReviewMode = $onSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::MODE];
        $endRulesMode      = $endRules[IWorkflow::UPDATE][IWorkflow::MODE];
        $autoApproveMode   = $autoApprove[IWorkflow::MODE];
        $countedVotesMode  = $countedVotes[IWorkflow::MODE];

        $this->logger->trace(
            get_class($this) . ": Global end_rules " . var_export($endRules, true)
        );
        $this->logger->trace(
            get_class($this) . ": Global on_submit " . var_export($onSubmit, true)
        );

        $specificWorkflow = $affectedProjects !== null
            ? $this->getMergedAffectedWorkflowRules($affectedProjects)
            : $this->getMergedChangeWorkflowRules($change, $rulesToCheck);
        if ($specificWorkflow) {
            if ($withReviewMode === IWorkflow::MODE_POLICY || $withoutReviewMode === IWorkflow::MODE_POLICY) {
                // If policy then merge global rule in if stricter
                $globalWf = new Workflow($p4Admin);
                $globalWf->setOnSubmit($onSubmit);
                if ($withReviewMode === IWorkflow::MODE_POLICY) {
                    $this->mergeWithReviewRule($globalWf, $specificWorkflow);
                }
                if ($withoutReviewMode === IWorkflow::MODE_POLICY) {
                    $this->mergeWithoutReviewRule($globalWf, $specificWorkflow);
                }
            }
            if ($endRulesMode === IWorkflow::MODE_POLICY) {
                $globalWf = new Workflow($p4Admin);
                $globalWf->setEndRules($endRules);
                $this->mergeUpdateRule($globalWf, $specificWorkflow);
            }
            if ($autoApproveMode === IWorkflow::MODE_POLICY) {
                $globalWf = new Workflow($p4Admin);
                $globalWf->setAutoApprove(array(IWorkflow::AUTO_APPROVE => $this->getGlobalAutoApprove($config)));
                static::mergeAutoApproveRule($globalWf, $specificWorkflow);
            }
            if ($countedVotesMode === IWorkflow::MODE_POLICY) {
                $globalWf = new Workflow($p4Admin);
                $globalWf->setCountedVotes($countedVotes);
                static::mergeCountedVotesRule($globalWf, $specificWorkflow);
            }
        }
        unset($onSubmit[IWorkflow::WITH_REVIEW][IWorkflow::MODE]);
        unset($onSubmit[IWorkflow::WITHOUT_REVIEW][IWorkflow::MODE]);
        unset($endRules[IWorkflow::UPDATE][IWorkflow::MODE]);
        unset($autoApprove[IWorkflow::MODE]);
        unset($countedVotes[IWorkflow::MODE]);
        return $specificWorkflow
            ? $specificWorkflow
            : array(IWorkflow::WORKFLOW_RULES =>
                array(
                    IWorkflow::ON_SUBMIT     => $onSubmit,
                    IWorkflow::END_RULES     => $endRules,
                    IWorkflow::AUTO_APPROVE  => $autoApprove,
                    IWorkflow::COUNTED_VOTES => $countedVotes
                )
            );
    }

    /**
     * Makes sure the change id is the correct format and that it exists
     * @param $changeId string the change id.
     * @return array
     * @throws \P4\Exception
     */
    private function validateChangeId($changeId)
    {
        $error  = null;
        $result = $this->buildResult(self::STATUS_OK);
        try {
            $change = Change::fetchById($changeId, $this->services->get('p4_admin'));
            // So we do not have to fetch the change again we will put it in the result
            $result[$changeId] = $change;
        } catch (NotFoundException $e) {
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $error = $e;
        }
        if ($error) {
            $translator = $this->services->get('translator');
            $result     = $this->buildResult(self::STATUS_BAD_CHANGE, $translator->t($error->getMessage()));
        }
        return $result;
    }

    /**
     * Helper to build a result to return
     * @param $status
     * @param null $message
     * @return array
     */
    private function buildResult($status, $message = null)
    {
        if ($message) {
            $messages = array($message);
        } else {
            $messages = array();
        }
        return array(self::KEY_STATUS => $status, self::KEY_MESSAGES => $messages);
    }

    /**
     * Gets the default values for a workflow according to global configuration settings.
     * @return array the default values
     * @throws \Application\Config\ConfigException
     * @throws \Exception
     */
    public function getDefaults()
    {
        return $this->getDefaultsFromConfig($this->services->get('config'));
    }

    /**
     * Get the defaults from the given Config file.
     *
     * @param ConfigManager $config   The config file that has been passed.
     *
     * @return array
     * @throws \Application\Config\ConfigException
     * @throws \Exception
     */
    public static function getDefaultsFromConfig($config)
    {
        return array(
            IWorkflow::WORKFLOW_RULES => array(
                IWorkflow::ON_SUBMIT      => static::getGlobalOnSubmit($config),
                IWorkflow::END_RULES      => static::getGlobalEndRules($config),
                IWorkflow::AUTO_APPROVE   => static::getGlobalAutoApprove($config),
                IWorkflow::COUNTED_VOTES  => static::getGlobalCountedVotes($config)
            ),
            IWorkflow::NAME           => '',
            IWorkflow::DESCRIPTION    => '',
            IWorkflow::OWNERS         => null,
            IWorkflow::SHARED         => false
        );
    }

    /**
     * Get the default Rule for a given Workflow rule.
     *
     * @param ConfigManager  $config The config that is being processed
     * @param string         $rule   The rule we want returned.
     *
     * @return mixed
     * @throws \Application\Config\ConfigException
     * @throws \Exception
     */
    public static function getDefaultRuleFromConfig($config, $rule)
    {
        $defaults = static::getDefaultsFromConfig($config);
        return $defaults[IWorkflow::WORKFLOW_RULES][$rule][IWorkflow::RULE];
    }

    /**
     * Get the Project or branch counted vote rule if set otherwise fall back to default.
     *
     * @param Project       $project  The project we are getting information from.
     * @param ConfigManager $config   The config we are processing.
     * @param Connection    $p4Admin  The P4Admin connection.
     * @param string        $branch   The branch Id that we want to get information from.
     *
     * @deprecated use getBranchRule(...) instead
     * @return mixed
     * @throws ConfigException
     * @throws Exception
     * @throws RecordNotFoundException
     * @throws \Exception
     */
    public static function getProjectBranchCountedVote($project, $config, $p4Admin, $branch = null)
    {
        return static::getBranchRule(
            IWorkflow::COUNTED_VOTES,
            array($project->getId()=>array($branch)),
            array($project->getId()=>$project)
        );
    }
}
